@@ -53,7 +53,7 @@ libcdr::CDRContentCollector::CDRContentCollector(libcdr::CDRParserState &ps, lib
   m_page(ps.m_pages[0]), m_pageIndex(0), m_currentFildId(0.0), m_currentOutlId(0), m_spnd(0),
   m_currentObjectLevel(0), m_currentGroupLevel(0), m_currentVectLevel(0), m_currentPageLevel(0),
   m_currentImage(), m_currentText(), m_currentTextOffsetX(0.0), m_currentTextOffsetY(0.0),
-  m_currentBBox(), m_currentPath(), m_currentTransform(), m_fillTransform(),
+  m_currentBBox(), m_currentPath(), m_currentTransforms(), m_fillTransforms(),
   m_polygon(0), m_isInPolygon(false), m_isInSpline(false), m_outputElements(0),
   m_contentOutputElements(), m_fillOutputElements(),
   m_groupLevels(), m_groupTransforms(), m_splineData(), m_fillOpacity(1.0), m_ps(ps)
@@ -124,7 +124,7 @@ void libcdr::CDRContentCollector::collectGroup(unsigned level)
   outputElement.addEndGroup();
   m_outputElements->push(outputElement);
   m_groupLevels.push(level);
-  m_groupTransforms.push(CDRTransform());
+  m_groupTransforms.push(CDRTransforms());
 }
 
 void libcdr::CDRContentCollector::collectVect(unsigned level)
@@ -209,6 +209,7 @@ void libcdr::CDRContentCollector::_flushCurrentPath()
     m_splineData.clear();
     m_isInSpline = false;
     bool firstPoint = true;
+    bool wasMove = false;
     double initialX = 0.0;
     double initialY = 0.0;
     double previousX = 0.0;
@@ -220,20 +221,25 @@ void libcdr::CDRContentCollector::_flushCurrentPath()
     _fillProperties(style, gradient);
     _lineProperties(style);
     outputElement.addStyle(style, gradient);
-    m_currentPath.transform(m_currentTransform);
+    m_currentPath.transform(m_currentTransforms);
     if (!m_groupTransforms.empty())
       m_currentPath.transform(m_groupTransforms.top());
     CDRTransform tmpTrafo(1.0, 0.0, -m_page.offsetX, 0.0, 1.0, -m_page.offsetY);
     m_currentPath.transform(tmpTrafo);
     tmpTrafo = CDRTransform(1.0, 0.0, 0.0, 0.0, -1.0, m_page.height);
     m_currentPath.transform(tmpTrafo);
-    WPXPropertyListVector tmpPath;
-    m_currentPath.writeOut(tmpPath);
+
+    std::vector<WPXPropertyList> tmpPath;
+
     WPXPropertyListVector path;
-    WPXPropertyListVector::Iter i(tmpPath);
-    bool ignoreM = false;
+    m_currentPath.writeOut(path);
+
+    bool isPathClosed = m_currentPath.isClosed();
+
+    WPXPropertyListVector::Iter i(path);
     for (i.rewind(); i.next();)
     {
+      bool ignoreM = false;
       if (!i()["libwpg:path-action"])
         continue;
       if (i()["svg:x"] && i()["svg:y"])
@@ -245,58 +251,95 @@ void libcdr::CDRContentCollector::_flushCurrentPath()
           initialX = x;
           initialY = y;
           firstPoint = false;
+          wasMove = true;
         }
         else if (i()["libwpg:path-action"]->getStr() == "M")
         {
+          // This is needed for a good generation of path from polygon
           if (CDR_ALMOST_ZERO(previousX - x) && CDR_ALMOST_ZERO(previousY - y))
             ignoreM = true;
           else
           {
-            if ((CDR_ALMOST_ZERO(initialX - previousX) && CDR_ALMOST_ZERO(initialY - previousY)) || (style["draw:fill"] && style["draw:fill"]->getStr() != "none"))
+            if (!tmpPath.empty())
             {
-              WPXPropertyList node;
-              node.insert("libwpg:path-action", "Z");
-              path.append(node);
+              if (!wasMove)
+              {
+                if ((CDR_ALMOST_ZERO(initialX - previousX) && CDR_ALMOST_ZERO(initialY - previousY)) || isPathClosed)
+                {
+                  WPXPropertyList node;
+                  node.insert("libwpg:path-action", "Z");
+                  tmpPath.push_back(node);
+                }
+              }
+              else
+              {
+                tmpPath.pop_back();
+              }
             }
+          }
+
+          if (!ignoreM)
+          {
             initialX = x;
             initialY = y;
+            wasMove = true;
           }
+
+        }
+        else
+          wasMove = false;
+
+        if (!ignoreM)
+        {
+          tmpPath.push_back(i());
+          previousX = x;
+          previousY = y;
+        }
+
+      }
+    }
+    if (!tmpPath.empty())
+    {
+      if (!wasMove)
+      {
+        if ((CDR_ALMOST_ZERO(initialX - previousX) && CDR_ALMOST_ZERO(initialY - previousY)) || isPathClosed)
+        {
+          WPXPropertyList closedPath;
+          closedPath.insert("libwpg:path-action", "Z");
+          tmpPath.push_back(closedPath);
         }
       }
-      previousX = x;
-      previousY = y;
-      if (!ignoreM)
-        path.append(i());
-      ignoreM = false;
+      else
+        tmpPath.pop_back();
     }
-    if ((CDR_ALMOST_ZERO(initialX - previousX) && CDR_ALMOST_ZERO(initialY - previousY)) || (style["draw:fill"] && style["draw:fill"]->getStr() != "none"))
+    if (!tmpPath.empty())
     {
-      WPXPropertyList node;
-      node.insert("libwpg:path-action", "Z");
-      path.append(node);
-    }
+      WPXPropertyListVector outputPath;
+      for (std::vector<WPXPropertyList>::const_iterator iter = tmpPath.begin(); iter != tmpPath.end(); ++iter)
+        outputPath.append(*iter);
 
-    outputElement.addPath(path);
+      outputElement.addPath(outputPath);
+
+    }
     m_currentPath.clear();
   }
+
   if (m_currentImage.getImage().size())
   {
     double cx = m_currentImage.getMiddleX();
     double cy = m_currentImage.getMiddleY();
-    m_currentTransform.applyToPoint(cx, cy);
+    m_currentTransforms.applyToPoint(cx, cy);
     if (!m_groupTransforms.empty())
       m_groupTransforms.top().applyToPoint(cx, cy);
     CDRTransform tmpTrafo(1.0, 0.0, -m_page.offsetX, 0.0, 1.0, -m_page.offsetY);
     tmpTrafo.applyToPoint(cx, cy);
     tmpTrafo = CDRTransform(1.0, 0.0, 0.0, 0.0, -1.0, m_page.height);
     tmpTrafo.applyToPoint(cx, cy);
-    double scaleX = (m_currentTransform.m_v0 < 0.0 ? -1.0 : 1.0)*sqrt(m_currentTransform.m_v0 * m_currentTransform.m_v0 + m_currentTransform.m_v1 * m_currentTransform.m_v1);
-    double scaleY = (m_currentTransform.m_v3 < 0.0 ? -1.0 : 1.0)*sqrt(m_currentTransform.m_v3 * m_currentTransform.m_v3 + m_currentTransform.m_v4 * m_currentTransform.m_v4);
-    bool flipX(scaleX < 0);
-    bool flipY(scaleY < 0);
-    double width = fabs(scaleX)*m_currentImage.getWidth();
-    double height = fabs(scaleY)*m_currentImage.getHeight();
-    double rotate = -atan2(m_currentTransform.m_v3, m_currentTransform.m_v4);
+    bool flipX(m_currentTransforms.getFlipX());
+    bool flipY(m_currentTransforms.getFlipY());
+    double width = m_currentTransforms.getScaleX() * m_currentImage.getWidth();
+    double height = m_currentTransforms.getScaleY() * m_currentImage.getHeight();
+    double rotate = m_currentTransforms.getRotation();
 
     WPXPropertyList propList;
 
@@ -332,7 +375,7 @@ void libcdr::CDRContentCollector::_flushCurrentPath()
   {
     double currentTextOffsetX = 0.0;
     double currentTextOffsetY = 0.0;
-    m_currentTransform.applyToPoint(currentTextOffsetX, currentTextOffsetY);
+    m_currentTransforms.applyToPoint(currentTextOffsetX, currentTextOffsetY);
     if (!m_groupTransforms.empty())
       m_groupTransforms.top().applyToPoint(currentTextOffsetX, currentTextOffsetY);
     WPXPropertyList textFrameProps;
@@ -370,23 +413,23 @@ void libcdr::CDRContentCollector::_flushCurrentPath()
   m_currentImage = libcdr::CDRImage();
   if (!outputElement.empty())
     m_outputElements->push(outputElement);
-  m_currentTransform = libcdr::CDRTransform();
-  m_fillTransform = libcdr::CDRTransform();
+  m_currentTransforms.clear();
+  m_fillTransforms = libcdr::CDRTransforms();
   m_fillOpacity = 1.0;
   m_currentText = CDRText();
 }
 
-void libcdr::CDRContentCollector::collectTransform(double v0, double v1, double x0, double v3, double v4, double y0, bool considerGroupTransform)
+void libcdr::CDRContentCollector::collectTransform(const CDRTransforms &transforms, bool considerGroupTransform)
 {
   if (m_currentObjectLevel)
-    m_currentTransform = libcdr::CDRTransform(v0, v1, x0, v3, v4, y0);
+    m_currentTransforms = transforms;
   else if (!m_groupLevels.empty() && considerGroupTransform)
-    m_groupTransforms.top() = CDRTransform(v0, v1, x0, v3, v4, y0);
+    m_groupTransforms.top() = transforms;
 }
 
-void libcdr::CDRContentCollector::collectFillTransform(double v0, double v1, double x0, double v3, double v4, double y0)
+void libcdr::CDRContentCollector::collectFillTransform(const CDRTransforms &fillTrafos)
 {
-  m_fillTransform = libcdr::CDRTransform(v0, v1, x0, v3, v4, y0);
+  m_fillTransforms = fillTrafos;
 }
 
 void libcdr::CDRContentCollector::collectLevel(unsigned level)
@@ -637,8 +680,8 @@ void libcdr::CDRContentCollector::_fillProperties(WPXPropertyList &propList, WPX
             double scaleY = 1.0;
             if (iter->second.imageFill.flags & 0x04) // scale fill with image
             {
-              scaleX = sqrt(m_currentTransform.m_v0 * m_currentTransform.m_v0 + m_currentTransform.m_v1 * m_currentTransform.m_v1);
-              scaleY = sqrt(m_currentTransform.m_v3 * m_currentTransform.m_v3 + m_currentTransform.m_v4 * m_currentTransform.m_v4);
+              scaleX = m_currentTransforms.getScaleX();
+              scaleY = m_currentTransforms.getScaleY();
             }
             propList.insert("svg:width", iter->second.imageFill.width * scaleX);
             propList.insert("svg:height", iter->second.imageFill.height * scaleY);
@@ -648,23 +691,23 @@ void libcdr::CDRContentCollector::_fillProperties(WPXPropertyList &propList, WPX
           {
             if (iter->second.imageFill.xOffset != 0.0 && iter->second.imageFill.xOffset != 1.0)
               propList.insert("draw:fill-image-ref-point-x", iter->second.imageFill.xOffset, WPX_PERCENT);
-            if (iter->second.imageFill.yOffset != 0.0 && iter->second.imageFill.yOffset != 0.0)
+            if (iter->second.imageFill.yOffset != 0.0 && iter->second.imageFill.yOffset != 1.0)
               propList.insert("draw:fill-image-ref-point-y", iter->second.imageFill.yOffset, WPX_PERCENT);
           }
           else
           {
-            if (m_fillTransform.m_x0 != 0.0)
+            if (m_fillTransforms.getTranslateX() != 0.0)
             {
-              double xOffset = m_fillTransform.m_x0 / iter->second.imageFill.width;
+              double xOffset = m_fillTransforms.getTranslateX() / iter->second.imageFill.width;
               while (xOffset < 0.0)
                 xOffset += 1.0;
               while (xOffset > 1.0)
                 xOffset -= 1.0;
               propList.insert("draw:fill-image-ref-point-x", xOffset, WPX_PERCENT);
             }
-            if (m_fillTransform.m_y0 != 0.0)
+            if (m_fillTransforms.getTranslateY() != 0.0)
             {
-              double yOffset = m_fillTransform.m_y0 / iter->second.imageFill.width;
+              double yOffset = m_fillTransforms.getTranslateY() / iter->second.imageFill.width;
               while (yOffset < 0.0)
                 yOffset += 1.0;
               while (yOffset > 1.0)
@@ -703,8 +746,8 @@ void libcdr::CDRContentCollector::_fillProperties(WPXPropertyList &propList, WPX
             double scaleY = 1.0;
             if (iter->second.imageFill.flags & 0x04) // scale fill with image
             {
-              scaleX = sqrt(m_currentTransform.m_v0 * m_currentTransform.m_v0 + m_currentTransform.m_v1 * m_currentTransform.m_v1);
-              scaleY = sqrt(m_currentTransform.m_v3 * m_currentTransform.m_v3 + m_currentTransform.m_v4 * m_currentTransform.m_v4);
+              scaleX = m_currentTransforms.getScaleX();
+              scaleY = m_currentTransforms.getScaleY();
             }
             propList.insert("svg:width", iter->second.imageFill.width * scaleX);
             propList.insert("svg:height", iter->second.imageFill.height * scaleY);
@@ -714,23 +757,23 @@ void libcdr::CDRContentCollector::_fillProperties(WPXPropertyList &propList, WPX
           {
             if (iter->second.imageFill.xOffset != 0.0 && iter->second.imageFill.xOffset != 1.0)
               propList.insert("draw:fill-image-ref-point-x", iter->second.imageFill.xOffset, WPX_PERCENT);
-            if (iter->second.imageFill.yOffset != 0.0 && iter->second.imageFill.yOffset != 0.0)
+            if (iter->second.imageFill.yOffset != 0.0 && iter->second.imageFill.yOffset != 1.0)
               propList.insert("draw:fill-image-ref-point-y", iter->second.imageFill.yOffset, WPX_PERCENT);
           }
           else
           {
-            if (m_fillTransform.m_x0 != 0.0)
+            if (m_fillTransforms.getTranslateX() != 0.0)
             {
-              double xOffset = m_fillTransform.m_x0 / iter->second.imageFill.width;
+              double xOffset = m_fillTransforms.getTranslateX() / iter->second.imageFill.width;
               while (xOffset < 0.0)
                 xOffset += 1.0;
               while (xOffset > 1.0)
                 xOffset -= 1.0;
               propList.insert("draw:fill-image-ref-point-x", xOffset, WPX_PERCENT);
             }
-            if (m_fillTransform.m_y0 != 0.0)
+            if (m_fillTransforms.getTranslateY() != 0.0)
             {
-              double yOffset = m_fillTransform.m_y0 / iter->second.imageFill.width;
+              double yOffset = m_fillTransforms.getTranslateY() / iter->second.imageFill.width;
               while (yOffset < 0.0)
                 yOffset += 1.0;
               while (yOffset > 1.0)
@@ -763,8 +806,8 @@ void libcdr::CDRContentCollector::_fillProperties(WPXPropertyList &propList, WPX
             double scaleY = 1.0;
             if (iter->second.imageFill.flags & 0x04) // scale fill with image
             {
-              scaleX = sqrt(m_currentTransform.m_v0 * m_currentTransform.m_v0 + m_currentTransform.m_v1 * m_currentTransform.m_v1);
-              scaleY = sqrt(m_currentTransform.m_v3 * m_currentTransform.m_v3 + m_currentTransform.m_v4 * m_currentTransform.m_v4);
+              scaleX = m_currentTransforms.getScaleX();
+              scaleY = m_currentTransforms.getScaleY();
             }
             propList.insert("svg:width", iter->second.imageFill.width * scaleX);
             propList.insert("svg:height", iter->second.imageFill.height * scaleY);
@@ -774,23 +817,23 @@ void libcdr::CDRContentCollector::_fillProperties(WPXPropertyList &propList, WPX
           {
             if (iter->second.imageFill.xOffset != 0.0 && iter->second.imageFill.xOffset != 1.0)
               propList.insert("draw:fill-image-ref-point-x", iter->second.imageFill.xOffset, WPX_PERCENT);
-            if (iter->second.imageFill.yOffset != 0.0 && iter->second.imageFill.yOffset != 0.0)
+            if (iter->second.imageFill.yOffset != 0.0 && iter->second.imageFill.yOffset != 1.0)
               propList.insert("draw:fill-image-ref-point-y", iter->second.imageFill.yOffset, WPX_PERCENT);
           }
           else
           {
-            if (m_fillTransform.m_x0 != 0.0)
+            if (m_fillTransforms.getTranslateX() != 0.0)
             {
-              double xOffset = m_fillTransform.m_x0 / iter->second.imageFill.width;
+              double xOffset = m_fillTransforms.getTranslateX() / iter->second.imageFill.width;
               while (xOffset < 0.0)
                 xOffset += 1.0;
               while (xOffset > 1.0)
                 xOffset -= 1.0;
               propList.insert("draw:fill-image-ref-point-x", xOffset, WPX_PERCENT);
             }
-            if (m_fillTransform.m_y0 != 0.0)
+            if (m_fillTransforms.getTranslateY() != 0.0)
             {
-              double yOffset = m_fillTransform.m_y0 / iter->second.imageFill.width;
+              double yOffset = m_fillTransforms.getTranslateY() / iter->second.imageFill.width;
               while (yOffset < 0.0)
                 yOffset += 1.0;
               while (yOffset > 1.0)
@@ -839,8 +882,8 @@ void libcdr::CDRContentCollector::_lineProperties(WPXPropertyList &propList)
       double scale = 1.0;
       if (iter->second.lineType & 0x20) // scale line with image
       {
-        scale = sqrt(m_currentTransform.m_v0 * m_currentTransform.m_v0 + m_currentTransform.m_v1 * m_currentTransform.m_v1);
-        double scaleY = sqrt(m_currentTransform.m_v3 * m_currentTransform.m_v3 + m_currentTransform.m_v4 * m_currentTransform.m_v4);
+        scale = m_currentTransforms.getScaleX();
+        double scaleY = m_currentTransforms.getScaleY();
         if (scaleY > scale)
           scale = scaleY;
       }
@@ -1021,7 +1064,7 @@ void libcdr::CDRContentCollector::collectFillOpacity(double opacity)
 void libcdr::CDRContentCollector::collectBBox(double x0, double y0, double x1, double y1)
 {
   CDRBBox bBox(x0, y0, x1, y1);
-  if (m_currentVectLevel && m_page.width == 0.0 && m_page.width == 0.0)
+  if (m_currentVectLevel && m_page.width == 0.0 && m_page.height == 0.0)
   {
     m_page.width = bBox.getWidth();
     m_page.height = bBox.getHeight();
